/* Gtick  -  A Gtk+ Metronome
 *
 *   dsp.c: /dev/dsp access code
 *
 * Copyright (c) 1999, Alex Roberts
 * Copyright (c) 2003, Roland Stigge <stigge@antcom.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
 /* Funky /dev/dsp programming stuff here.. */

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef FREEBSD
#include <machine/soundcard.h>
#else
#include <linux/soundcard.h>
#endif
#include <glib.h>

#include "gtick.h"
#include "tickdata.c"

static int dsp;
static int rate, channels, format;
static int fragmentsize;

static gint input;

static unsigned char *fragment;
static unsigned char *tickdata2;
      
int gtInitDSP (TikTok *ticker)
{
  	int i;
	
	fragmentsize = 0x7fff000E;

#ifdef DEBUG
	g_print ("gtInitDSP: Initialising /dev/dsp..\n");
#endif

	/* Initialise /dev/dsp */
	if ((dsp = open ("/dev/dsp", O_WRONLY)) == -1)
	  {
	    perror ("/dev/dsp");
	    return -1;
	  }

	if (ioctl(dsp, SNDCTL_DSP_SETFRAGMENT, &fragmentsize) == -1 ) {
	  perror("SNDCTL_DSP_SETFRAGMENT");
	}

	/* Set up Output format */
	format = AFMT_U8; /* AFMT_S16_LE */
	if (ioctl (dsp, SNDCTL_DSP_SETFMT, &format) == -1)  
	  { /* Fatal error */  
	    perror ("SNDCTL_DSP_SETFMT");  
	    exit(1);  
	  }        
                        
	/* Set dsp to Mono */
	channels = 1;
	if (ioctl (dsp, SNDCTL_DSP_CHANNELS, &channels) == -1) {
	  perror("SNDCTL_DSP_CHANNELS");
	  /* exit(1); */
	}
	
	/* Set the DSP rate (in Hz) */
	rate = 44100;
	if (ioctl (dsp, SNDCTL_DSP_SPEED, &rate) == -1) {
	  perror("SNDCTL_DSP_SPEED");
	  /* exit(1); */
	}

	/* "verify" fragment size:
	*  let the driver actually calculate the fragment size
	*/
	if (ioctl(dsp, SNDCTL_DSP_GETBLKSIZE, &fragmentsize)) {
	  perror("SNDCTL_DSP_GETBLKSIZE");
	}

#ifdef DEBUG
	g_print ("fragment size: %d\n", fragmentsize);
#endif

	if (!(fragment = g_malloc(fragmentsize))) {
	  perror("g_malloc");
	}

	/* generate second ticks */
	if (!(tickdata2 = g_malloc(sizeof(tickdata)))) {
	  perror("g_malloc");
	}
	for (i = 0; i < sizeof(tickdata); i++) {
	  tickdata2[i] =
	    (unsigned char) (((int)tickdata[i] - 128) * 1 / 2 + 128);
	}

	input = gdk_input_add(dsp, GDK_INPUT_WRITE, gtDSPFeed, ticker);

	return 0;
}

void gtCloseDSP ()
{
	if (ioctl(dsp, SNDCTL_DSP_RESET, 0) == -1) {
	  perror("SNDCTL_DSP_RESET");
	}
	
	gdk_input_remove(input);
	
  	g_free(fragment);
	g_free(tickdata2);

#ifdef DEBUG
	g_print ("gtCloseDSP: Closing /dev/dsp..\n");
#endif
	close (dsp);
}

/*
 * Feed dsp device with next samples
 *
 * used as output start and callback
 */
void gtDSPFeed(gpointer data, gint fd, GdkInputCondition condition)
{
  static unsigned char v = 128;
  static int cyclepos = 0;
  
  audio_buf_info info;
  TikTok *ticker = (TikTok *)data;
  int ticklen = rate * ticker->period / 1000;
  int cyclelen = ticker->count * ticklen;
  unsigned char *thedata;
  int index;
 
  /* get number of fragments to write to dsp */
  if (ioctl(dsp, SNDCTL_DSP_GETOSPACE, &info) == -1) {
    perror("SNDCTL_DSP_GETOSPACE");
  }

  if (!ticker->ticking) cyclepos = 0;
  if (cyclepos >= cyclelen) cyclepos = 0;

  /* write as much fragments as possible */
  while (info.fragments) {
    int i;
    
    
    /* generate fragment */
    for (i = 0; i < fragmentsize; i++) {
      if (ticker->ticking) { /* metronome on */
	
	if ((index = cyclepos % ticklen) < sizeof(tickdata)) { /* tick! */
	  if (cyclepos / ticklen) /* 2nd tick or beyond */
	    thedata = tickdata2;
	  else /* 1st tick */
	    thedata = tickdata;
	  fragment[i] = thedata[index];
	  
	} else { /* silence (between ticks)  */
	  fragment[i] = v;
	}
	  
        cyclepos ++;
        if (cyclepos >= cyclelen) cyclepos = 0;
	
      } else { /* silence (metronome off) */
        fragment[i] = v;
      }
    }
    
    write(dsp, fragment, fragmentsize);
    info.fragments--;
  }
}

#ifdef TEST_DSP
int main ()
{
int t;

	gtInitDSP ();
  
	for (t = 0; t < 6; t++)
	   gtDSPTick ();

	gtCloseDSP ();
	
	return 0;
}
#endif /* TEST_DSP */
